# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2024 Joe Reynolds and Mark Liddell
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""BIFROST-HTTr CLI - Unified command-line interface for BIFROST-HTTr package.

This module provides a unified command-line interface for the BIFROST-HTTr package,
organizing commands into logical groups for analysis, reporting, and data management.
"""

from pathlib import Path
from typing import (  # noqa: F401 - used in type hints
    Any,
    Optional,
    Union,
)

import click
import numpy as np
import pandas as pd

from .core.analysis import run_concentration_response_analysis
from .core.data_processing import (
    generate_bifrost_inputs,
    process_batches,
    process_data,
    validate_config,
    validate_counts_table,
    validate_meta_data,
    validate_output_directory,
    validate_substances_cell_types,
)
from .core.model import compile_stan_model
from .utils.compression import compress_output
from .utils.config import convert_meta_data, load_yaml_file
from .utils.logging import get_logger
from .visualization.data import BifrostData
from .visualization.report import BifrostMultiQCReport

# Configure logging automatically when getting logger
logger = get_logger(__name__)


@click.group(name="bifrost-httr")
@click.version_option(package_name="bifrost-httr")
def cli() -> None:
    """BIFROST-HTTr Analysis - Bayesian inference for region of signal threshold."""


@click.command()
@click.option(
    "--data-files",
    "-f",
    multiple=True,
    help="List of probe .pkl files to process",
)
@click.option(
    "--model-executable",
    "-m",
    type=click.Path(exists=True),
    help="Path to compiled Stan model executable (optional, will use default model if not provided)",
)
@click.option(
    "--n-cores",
    "-n",
    type=int,
    help="Number of cores to use in multiprocessing",
)
@click.option("--seed", "-s", type=int, help="Optional random seed for reproducibility")
def run_analysis(
    data_files: tuple[str, ...],
    model_executable: str | None,
    n_cores: int | None,
    seed: int | None,
) -> None:
    """Run concentration-response analysis on BIFROST data files."""
    if not data_files:
        click.echo("Error: No data files specified", err=True)
        return

    # If no model executable is provided, compile the default model
    if model_executable is None:
        model_executable = compile_stan_model()
    else:
        model_executable = Path(model_executable)

    run_concentration_response_analysis(
        files_to_process=data_files,
        model_executable=model_executable,
        number_of_cores=n_cores or 1,
        seed=seed,
    )


@click.command()
@click.option(
    "--summary-file",
    "-s",
    required=True,
    type=click.Path(exists=True),
    help="Path to the summary JSON file",
)
@click.option(
    "--test-substance",
    "-t",
    default="MyChemical",
    help="Name of the test substance",
)
@click.option(
    "--cell-type",
    "-c",
    default="MyCell",
    help="Type of cell used in the test",
)
@click.option(
    "--output-name",
    "-o",
    default="multiqc_report.html",
    help="Name for the output report",
)
@click.option(
    "--timepoint",
    default="24 hours",
    help="Exposure duration within experiment",
)
@click.option(
    "--conc-units",
    type=click.Choice(["uM", "ugml-1", "mgml-1"]),
    default="uM",
    help="Concentration units",
)
@click.option(
    "--interactive-plots",
    is_flag=True,
    help="Force interactive plots (may be faster for large datasets)",
)
@click.option(
    "--n-fold-change-probes",
    type=int,
    default=5,
    help="Number of most up/down regulated probes to show",
)
@click.option(
    "--cds-threshold",
    type=float,
    default=0.5,
    help="Concentration-Dependency Score threshold for filtering probes",
)
@click.option(
    "--n-lowest-means",
    type=int,
    default=10,
    help="Number of lowest mean PoD probes to show",
)
@click.option(
    "--n-pod-stats",
    type=int,
    default=100,
    help="Number of probes to include in PoD statistics table",
)
@click.option(
    "--plot-height",
    type=int,
    default=400,
    help="Height of concentration-response plots in pixels",
)
@click.option(
    "--pod-vs-fc-height",
    type=int,
    default=600,
    help="Height of PoD vs Fold Change plot in pixels",
)
@click.option(
    "--no-cds-threshold",
    is_flag=True,
    help="Do not filter probes by CDS threshold in summary tables and lowest mean PoDs section",
)
@click.option(
    "--custom-templates",
    type=click.Path(exists=True),
    help="Path to custom template YAML file to override default report templates",
)
def create_report(
    summary_file: str,
    test_substance: str,
    cell_type: str,
    output_name: str,
    timepoint: str,
    conc_units: str,
    n_fold_change_probes: int,
    cds_threshold: float,
    n_lowest_means: int,
    n_pod_stats: int,
    plot_height: int,
    pod_vs_fc_height: int,
    custom_templates: str | None,
    *,
    interactive_plots: bool,
    no_cds_threshold: bool,
) -> None:
    """Create Bifrost HTTR reports using MultiQC."""
    data = BifrostData(summary_file)
    report = BifrostMultiQCReport(
        data,
        test_substance=test_substance,
        cell_type=cell_type,
        output_name=output_name,
        timepoint=timepoint,
        conc_units=conc_units,
        interactive_plots=interactive_plots,
        n_fold_change_probes=n_fold_change_probes,
        cds_threshold=cds_threshold,
        n_lowest_means=n_lowest_means,
        n_pod_stats=n_pod_stats,
        plot_height=plot_height,
        pod_vs_fc_height=pod_vs_fc_height,
        no_cds_threshold=no_cds_threshold,
        custom_templates=custom_templates,
    )
    report.create_report()


@click.command()
@click.option(
    "--fits-dir",
    "-f",
    required=True,
    type=click.Path(exists=True),
    help="Directory containing probe .pkl files to process",
)
@click.option(
    "--output",
    "-o",
    required=True,
    type=click.Path(),
    help="Path to the output json",
)
@click.option("--seed", "-s", type=int, help="Optional random seed for reproducibility")
@click.option(
    "--no-compression",
    is_flag=True,
    help="Save output as plain JSON without compression",
)
def compress_output_cmd(
    fits_dir: str,
    output: str,
    seed: int | None,
    *,
    no_compression: bool,
) -> None:
    """Compress intermediate output into a single pandas DataFrame."""
    compress_output(fits_dir, output, seed, no_compression=no_compression)


@click.command()
@click.option(
    "--meta-data",
    "-m",
    required=True,
    type=click.Path(exists=True),
    help="Path to meta data CSV file",
)
@click.option(
    "--meta-mapper",
    required=True,
    type=click.Path(exists=True),
    help="Path to meta data mapper YAML file",
)
@click.option(
    "--counts",
    "-c",
    required=True,
    type=click.Path(exists=True),
    help="Path to counts CSV file",
)
@click.option(
    "--substances-cell-types",
    required=True,
    type=click.Path(exists=True),
    help="Path to substances and cell types YAML file",
)
@click.option(
    "--batch-key",
    default="Batch",
    help="Field to use as batch key in the BIFROST model",
)
@click.option(
    "--min-percent-mapped-reads",
    type=float,
    default=80.0,
    help="Minimum percentage of mapped reads required",
)
@click.option(
    "--min-num-mapped-reads",
    type=int,
    default=1000000,
    help="Minimum number of mapped reads required",
)
@click.option(
    "--min-avg-treatment-count",
    type=int,
    default=10,
    help="Minimum average treatment count required",
)
@click.option(
    "--output-dir",
    "-o",
    required=True,
    type=click.Path(),
    help="Directory to store outputs",
)
@click.option(
    "--test-probes",
    type=int,
    help="Number of probes to sample for testing (optional)",
)
@click.option(
    "--random-seed",
    type=int,
    default=5,
    help="Random seed for reproducible probe selection",
)
def prepare_inputs(
    meta_data: str,
    meta_mapper: str,
    counts: str,
    substances_cell_types: str,
    batch_key: str,
    min_percent_mapped_reads: float,
    min_num_mapped_reads: int,
    min_avg_treatment_count: int,
    output_dir: str,
    test_probes: int | None,
    random_seed: int,
) -> None:
    """Prepare Bifrost inputs from meta data and counts."""
    # Set random seed for reproducibility
    np.random.seed(random_seed)

    # Load meta data file and convert in to common format
    meta_raw = pd.read_csv(meta_data)
    meta_data_mapper = load_yaml_file(meta_mapper)
    meta = convert_meta_data(meta_raw, meta_data_mapper)
    validate_meta_data(meta)

    # Load counts table
    counts_table = pd.read_csv(counts)
    validate_counts_table(counts_table)

    # Load substances and cell types
    substances_cell_types_dict = load_yaml_file(substances_cell_types)
    validate_substances_cell_types(substances_cell_types_dict)

    # Create config dictionary from arguments (CLI overrides YAML)
    config_dict = {
        "Test substances": substances_cell_types_dict["Test substances"],
        "Cell types": substances_cell_types_dict["Cell types"],
        "Additional divider": substances_cell_types_dict.get(
            "Additional divider",
            "N/A",
        ),
        "Specific filters": substances_cell_types_dict.get("Specific filters", None),
        "Batch key": batch_key,
        "Minimum percent mapped reads": min_percent_mapped_reads,
        "Minimum number mapped reads": min_num_mapped_reads,
        "Minimum average treatment count": min_avg_treatment_count,
    }
    validate_config(config_dict)

    # Make directory for storing outputs
    output_dir = Path(output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)
    generate_bifrost_inputs(meta, counts_table, config_dict, output_dir)

    # Reduce datasets to specified number of probes if test_probes is provided
    if test_probes and test_probes > 0:
        for i in output_dir.iterdir():
            file_path = output_dir / i.name
            if file_path.suffix != ".json":
                continue
            df = pd.read_json(file_path, orient="index", typ="series")
            click.echo(f"Sampling {test_probes} probes from {i.name}")
            index = np.random.choice(len(df["probes"]), size=test_probes, replace=False)
            df["probes"] = np.array(df["probes"])[index]
            df["counts"] = np.array(df["counts"], dtype="int")[index]
            df.to_json(file_path, orient="index")


@click.command()
@click.option(
    "--input-file",
    "-i",
    required=True,
    type=click.Path(exists=True),
    help="Path to input data json",
)
@click.option(
    "--analysis-dir",
    "-a",
    required=True,
    type=click.Path(),
    help="Path to analysis directory",
)
@click.option(
    "--batch-size",
    "-b",
    type=int,
    default=0,
    help="Number of files per batch",
)
@click.option(
    "--batch-mode",
    type=click.Choice(["batch", "all"]),
    default="all",
    help='Batch mode: "batch" for individual archives, "all" for single archive',
)
@click.option("--prefix", "-p", required=True, help="Prefix for output files")
@click.option(
    "--archive-mode",
    type=click.Choice(["tar", "directory"]),
    default="tar",
    help='Archive mode: "tar" for .tar.gz archives (default), "directory" for folders of files',
)
def split_data(
    input_file: str,
    analysis_dir: str,
    batch_size: int,
    batch_mode: str,
    prefix: str,
    archive_mode: str,
) -> None:
    """Split data into processing batches."""
    # Validate and create analysis directory
    analysis_dir = Path(analysis_dir)
    validate_output_directory(analysis_dir)
    analysis_dir.mkdir(parents=True, exist_ok=True)

    # Process the data into pickle files
    process_data(
        input_file_path=input_file,
        path_to_output=analysis_dir,
        testing_mode=False,
    )

    # Process the pickle files into batches
    data_dir = analysis_dir / "Data"
    manifest_file = process_batches(
        data_dir,
        prefix,
        batch_size,
        batch_mode,
        archive_mode,
    )
    click.echo(f"Created manifest file: {manifest_file}")


@click.command()
@click.argument("stan_file", type=click.Path(exists=True), required=False)
def compile_model(stan_file: str | None) -> None:
    """Compile a Stan model file (uses built-in model if no file specified)."""
    if stan_file is None:
        exe_file = compile_stan_model()
    else:
        exe_file = compile_stan_model(Path(stan_file))
    click.echo(f"Model compiled successfully: {exe_file}")


# Register all commands
cli.add_command(run_analysis)
cli.add_command(create_report)
cli.add_command(compress_output_cmd)
cli.add_command(prepare_inputs)
cli.add_command(split_data)
cli.add_command(compile_model)

if __name__ == "__main__":
    cli()
